#ifndef BBCmd_BBCMD_P_H
#define BBCmd_BBCMD_P_H

/**
 * This is the private header file of BBCmd.
 * The modules that have to know the internal of BBCmd
 * (e.g. defining its original command or other reason)
 * may include this.
 */

#include "BBCmd.h"




/* BBCmdCommand */

#define BBCmdCommand_HEAD				\
	const BBCmdCommandDescriptor* cmd_descriptor;

#define BBCmdCommand_INIT(cmd,desc)		\
	( ((cmd)->cmd_descriptor = (desc)) )

struct _BBCmdCommand {
	BBCmdCommand_HEAD

	/* cmd_descriptor: the descriptor representing a command */
};




/* BBCmdCommandDescriptor */

#define BBCmdCommandDescriptor_HEAD				\
	const char* cmd_name;					\
	const char* const* cmd_alias;				\
	const char* cmd_doc;					\
	int cmd_size;						\
	const char* (*cmd_initialize)( BBCmdCommand*,		\
	                               int, const char* const*,	\
	                               const BBCmdCompiler* );	\
	void (*cmd_del)(BBCmdCommand*);				\
	int (*cmd_execute)(const BBCmdCommand*);

struct _BBCmdCommandDescriptor {
	BBCmdCommandDescriptor_HEAD

	/**
	 * cmd_name
	 *     The name of a command.
	 *
	 * cmd_alias
	 *     The aliases of a command.
	 *     This member points a NULL terminated array
	 *     which contains the aliases.
	 *     This member can be NULL to represent that
	 *     the command has no alias.
	 *
	 * cmd_doc
	 *     The short document of a command.
	 *
	 * cmd_size
	 *     The size of an instance of a command (bytes).
	 *
	 * cmd_initialize
	 *     The function which initializes a given instance (1st arg).
	 *
	 *     This function must initialize only its members,
	 *     because the library will initialize BBCmdCommand's members
	 *     before calling this.
	 *
	 *     The 3rd arg is string array and the 2nd arg is number of it.
	 *     The 4th arg is the compiler that calls this method.
	 *
	 *     This function must return NULL for success
	 *     or a string for failure.
	 *     The string represents an error message.
	 *
	 *     The 0th element of the 3rd arg contains a command name.
	 *
	 *     If this function failed, the given instance will be
	 *     BBCmdCommand_Del immediately.
	 *     So take care bugs freeing uninitialized pointer or others.
	 * 
	 * cmd_del
	 *     The function which deletes a given instance.
	 *
	 *     This function must delete only its members,
	 *     because the library will delete BBCmdCommand's members
	 *     after calling this.
	 * 
	 * cmd_execute
	 *     The function which executes a given instance.
	 *
	 *     This function must return 0 for success
	 *     or positive number for failure.
	 *
	 *     This function must not return negative number.
	 *     The numbers are reserved for the library.
	 */
};




/* BBCmdCompiler */
struct _BBCmdCompiler {
	/* `cd' is short for `command definition' */
	const BBCmdCommandDescriptor** cd_table;
	int cd_table_size;
};




/**
 * Constants for returning value of cmd_execute.
 * (RC is short for `Return Code')
 */

#define BBCmd_RC_TRUE 0
#define BBCmd_RC_FALSE 1
#define BBCmd_RC_BREAK (-13)
#define BBCmd_RC_CONTINUE (-14)




#endif
