/* Implementation of broam, deskleft and others */

#include <assert.h>
#include <stdlib.h>
#include <string.h>

#define WIN32_LEAN_AND_MEAN
#include <windows.h>

#include <BBApi-C.h>

#include "Message_p.h"




struct _BBCmdMessageCommandParameter {
	const char* name;
	int num_of_args;
	int need_proxy;
	UINT message;
	WPARAM wparam;
	LPARAM lparam;
};

static const BBCmdMessageCommandParameter PARAMETER_TABLE[] = {
	{"aboutplugins",	0,  0, BB_ABOUTPLUGINS,	0, 0},
	{"aboutstyle",		0,  0, BB_ABOUTSTYLE,	0, 0},
	{"adddesk",		0,  0, BB_WORKSPACE,	2, 0},
	{"broam",		1, !0, BB_BROADCAST,	0, 0},
	{"deldesk",		0,  0, BB_WORKSPACE,	3, 0},
	{"deskleft",		0,  0, BB_WORKSPACE,	0, 0},
	{"deskright",		0,  0, BB_WORKSPACE,	1, 0},
	{"editblackbox",	0,  0, BB_EDITFILE,	4, 0},
	{"editextensions",	0,  0, BB_EDITFILE,	3, 0},
	{"editmenu",		0,  0, BB_EDITFILE,	1, 0},
	{"editplugins",		0,  0, BB_EDITFILE,	2, 0},
	{"editstyle",		0,  0, BB_EDITFILE,	0, 0},
	{"editworkspacename",	0,  0, BB_WORKSPACE,	13, 0},
	{"gatherwindows",	0,  0, BB_WORKSPACE,	5, 0},
	{"hibernate",		0,  0, BB_SHUTDOWN,	3, 0},
	{"hidemenu",		0,  0, BB_HIDEMENU,	0, 0},
	{"label",		1, !0, BB_SETTOOLBARLABEL, 0, 0},
	{"lockworkstation",	0,  0, BB_SHUTDOWN,	5, 0},
	{"logoff",		0,  0, BB_SHUTDOWN,	2, 0},
	{"mainmenu",		0,  0, BB_MENU,		0, 0},
	{"quit",		0,  0, BB_QUIT,		0, 0},
	{"reboot",		0,  0, BB_SHUTDOWN,	1, 0},
	{"reconfig",		0,  0, BB_RECONFIGURE,	0, 0},
	{"restart",		0,  0, BB_RESTART,	0, 0},
	{"run",			0,  0, BB_RUN,		0, 0},
	{"shutdown",		0,  0, BB_SHUTDOWN,	0, 0},
	{"slitmenu",		0,  0, BB_MENU,		4, 0},
	{"style",		1, !0, BB_SETSTYLE,	0, 0},
	{"stylesmenu",		0,  0, BB_MENU,		5, 0},
	{"suspend",		0,  0, BB_SHUTDOWN,	4, 0},
	{"systembarmenu",	0,  0, BB_MENU,		3, 0},
	{"toggleplugins",	0,  0, BB_TOGGLEPLUGINS,0, 0},
	{"toggleslit",		0,  0, BB_TOGGLESLIT,	0, 0},
	{"togglesystembar",	0,  0, BB_TOGGLETRAY,	0, 0},
	{"toggletoolbar",	0,  0, BB_TOGGLETOOLBAR,0, 0},
	{"toggletray",		0,  0, BB_TOGGLETRAY,	0, 0},
	{"toolbarmenu",		0,  0, BB_MENU,		2, 0},
	{"workspacemenu",	0,  0, BB_MENU,		1, 0},
	{NULL,			0,  0, 0,		0, 0}
};

static const BBCmdMessageCommandParameter*
search_parameter(const char* name)
{
	const BBCmdMessageCommandParameter* p;

	for (p = PARAMETER_TABLE; p->name != NULL; p++) {
		if (!strcmp(name, p->name))
			return p;
	}

	return NULL;
}

static const char*
replace_alias(const BBCmdCommandDescriptor* d, const char* alias)
{
	const char* const* i;

	if (!strcmp(d->cmd_name, alias)) return d->cmd_name;

	i = d->cmd_alias;
	if (i == NULL) return NULL;

	while (*i != NULL) {
		if (!strcmp(*i, alias)) return d->cmd_name;
		i++;
	}

	return NULL;
}




static const char*
_BBCmdMessageCommand_initialize( BBCmdCommand* _self,
                                 int ac, const char* const* av,
                                 const BBCmdCompiler* compiler )
{
	BBCmdMessageCommand* self;
	const BBCmdMessageCommandParameter* parameter;
	const char* name;

	/*assert(_self->cmd_descriptor == BBCmdMessageCommandDescriptor);*/
	self = (BBCmdMessageCommand*)_self;

	self->me_string = NULL;

	if (ac - 1 < 0) return "Too few arguments";

	name = replace_alias(self->cmd_descriptor, av[0]);
	if (name == NULL) return "Unable to replace alias (maybe BBCmd's bug)";

	parameter = search_parameter(name);
	if (parameter == NULL) return "Unknown command (maybe BBCmd's bug)";

	if (ac - 1 != parameter->num_of_args)
		return "Incorrect number of arguments";

	self->me_parameter = parameter;
	if (parameter->num_of_args != 0) {
		self->me_string = strdup(av[1]);
		if (self->me_string == NULL)
		  return "Out of memory (_BBCmdMessageCommand_initialize)";
	}

	return NULL;
}


static void
_BBCmdMessageCommand_del(BBCmdCommand* _self)
{
	BBCmdMessageCommand* self;

	/*assert(_self->cmd_descriptor == BBCmdMessageCommandDescriptor);*/
	self = (BBCmdMessageCommand*)_self;

	free(self->me_string);
}


static int
_BBCmdMessageCommand_execute(const BBCmdCommand* _self)
{
	BBCmdMessageCommand* self;
	HWND window;
	int rc;

	rc = BBCmd_RC_TRUE;

	/*assert(_self->cmd_descriptor == BBCmdMessageCommandDescriptor);*/
	self = (BBCmdMessageCommand*)_self;

	if (self->me_parameter->need_proxy) {
		int isspecial;
		char* swap = NULL;
		char* buf = NULL;

		isspecial = ( (self->me_parameter->message==BB_SETTOOLBARLABEL)
		              && (self->me_string[0] == '@') );
		if (isspecial) {
			HWND wWw;
			int length;

			wWw = FindWindow(self->me_string + 1, NULL);
			if (wWw == NULL) return BBCmd_RC_FALSE;

			length = GetWindowTextLength(wWw);
			if (length < 0) length = 0;

			buf = (char*)malloc(sizeof(char) * (length+1));
			if (buf == NULL) return BBCmd_RC_FALSE;

			GetWindowText(wWw, buf, length+1);

			swap = self->me_string;
			self->me_string = buf;
		}

		if (BBCmd_AmIPlugin()) {
			window = BBCmd_FindCore();
			if (window == NULL) {rc = BBCmd_RC_FALSE;  goto ERR;}
			SendMessage( window, self->me_parameter->message,
			             (WPARAM)NULL, (LPARAM)(self->me_string) );
		} else {
			COPYDATASTRUCT cds;

			window = BBCmd_FindProxy();
			if (window == NULL) {rc = BBCmd_RC_FALSE;  goto ERR;}

			cds.dwData = self->me_parameter->message;
			cds.cbData = sizeof(char)*(strlen(self->me_string) +1);
			cds.lpData = self->me_string;

			SendMessage( window, WM_COPYDATA,
			             (WPARAM)NULL, (LPARAM)(&cds) );
		}

ERR:
		if (isspecial) {
			free(buf);
			self->me_string = swap;
		}
	} else {
		window = BBCmd_FindCore();
		if (window == NULL) return BBCmd_RC_FALSE;

		PostMessage( window,
		             self->me_parameter->message,
		             self->me_parameter->wparam,
		             self->me_parameter->lparam );
	}

	return rc;
}




#define DEFINE_DESC(NAME,DOC,...)				\
	static const char* _BBCmd__##NAME##__Command_alias[] = {\
		__VA_ARGS__,					\
		NULL						\
	};							\
	static const BBCmdCommandDescriptor			\
			_BBCmd__##NAME##__CommandDescriptor={	\
		#NAME,						\
		_BBCmd__##NAME##__Command_alias,		\
		DOC,						\
		sizeof(BBCmdMessageCommand),			\
		_BBCmdMessageCommand_initialize,		\
		_BBCmdMessageCommand_del,			\
		_BBCmdMessageCommand_execute			\
	};							\
	const BBCmdCommandDescriptor* const BBCmd__##NAME##__CommandDescriptor\
		= &_BBCmd__##NAME##__CommandDescriptor;


DEFINE_DESC(aboutplugins,	"()\nShow dialog about loaded plugins.",NULL)
DEFINE_DESC(aboutstyle,		"()\nShow dialog about current style.",NULL)
DEFINE_DESC(adddesk,		"()\nAdd a workspace.",		NULL)
DEFINE_DESC(broam,		"(s)\nSend a specified bro@m `s'.",	NULL)
DEFINE_DESC(deldesk,		"()\nDelete last workspace.",		NULL)
DEFINE_DESC(deskleft,	"()\nMove workspace to left.", "LeftWorkspace")
DEFINE_DESC(deskright,	"()\nMove workspace to right.", "RightWorkspace")
DEFINE_DESC(editblackbox,	"()\nEdit `blackbox.rc'.",		NULL)
DEFINE_DESC(editextensions,	"()\nEdit `extensions.rc'.",		NULL)
DEFINE_DESC(editmenu,		"()\nEdit `menu.rc'.",		NULL)
DEFINE_DESC(editplugins,	"()\nEdit `plugins.rc'.",		NULL)
DEFINE_DESC(editstyle,		"()\nEdit current style.",		NULL)
DEFINE_DESC(editworkspacename,	"()\nEdit workspace name.",		NULL)
DEFINE_DESC(gatherwindows, "()\nGather windows in current workspace.",NULL)
DEFINE_DESC(hibernate,		"()\nHibernate the system.",		NULL)
DEFINE_DESC(hidemenu,		"()\nHide main menu.",		NULL)
DEFINE_DESC(label,		"(s)\nSet `s' as toolbar label.",	NULL)
DEFINE_DESC(lockworkstation,	"()\nLock the system.",		NULL)
DEFINE_DESC(logoff,		"()\nLog off the system.",		NULL)
DEFINE_DESC(mainmenu,		"()\nShow main menu.",		NULL)
DEFINE_DESC(quit,		"()\nQuit the shell.",		NULL)
DEFINE_DESC(reboot,		"()\nReboot the system.",		NULL)
DEFINE_DESC(reconfig,		"()\nReconfigure the system.",	NULL)
DEFINE_DESC(restart,		"()\nRestart the system.",		NULL)
DEFINE_DESC(run,		"()\nShow `run' dialog.",		NULL)
DEFINE_DESC(shutdown,		"()\nShutdown the system.",		NULL)
DEFINE_DESC(slitmenu,		"()\nShow slit menu.",		NULL)
DEFINE_DESC(style,		"(s)\nSet file `s' as style.",	NULL)
DEFINE_DESC(stylesmenu,		"()\nShow styles menu.",		NULL)
DEFINE_DESC(suspend,		"()\nSuspend the system.",		NULL)
DEFINE_DESC(systembarmenu,	"()\nShow systembar menu.",		NULL)
DEFINE_DESC(toggleplugins,	"()\nToggle plugins visibility.",	NULL)
DEFINE_DESC(toggleslit,		"()\nToggle slit visibility.",	NULL)
DEFINE_DESC(togglesystembar,	"()\nToggle systembar visibility.",	NULL)
DEFINE_DESC(toggletoolbar,	"()\nToggle toolbar visibility.",	NULL)
DEFINE_DESC(toggletray,		"()\nToggletray visibility.",		NULL)
DEFINE_DESC(toolbarmenu,	"()\nShow toolbar menu.",		NULL)
DEFINE_DESC(workspacemenu,	"()\nShow workspace menu.",		NULL)




/* __END__ */
