/* Implementation of `window_move' and other window-related command */

#include <assert.h>
#include <stdlib.h>
#include <string.h>

#define WIN32_LEAN_AND_MEAN
#include <windows.h>

#include <BBApi-C.h>

#include "Window_p.h"




/* ---------------- window_move ---------------- */

static const char*
_BBCmdWindowMoveCommand_initialize( BBCmdCommand* _self,
                                   int ac, const char* const* av,
                                   const BBCmdCompiler* compiler )
{
	BBCmdWindowMoveCommand* self;
	const char* const VALID_XBASE = "EWCAR";
	const char* const VALID_YBASE = "NSCAR";
	const char* xybase;
	const char* const* rest_av;
	int iswmt;

	assert(_self->cmd_descriptor == BBCmdWindowMoveCommandDescriptor);
	self = (BBCmdWindowMoveCommand*)_self;

	self->wm_class = NULL;
	self->wm_title = NULL;

	iswmt = !strcmp(av[0], "window_move_to");
	if (ac-1 < (iswmt ? 2 : 3) || (iswmt ? 4 : 5) < ac-1)
		return "Incorrect number of arguments";

	xybase = (iswmt ? "RR" : av[1]);
	rest_av = (iswmt ? av + 1 : av + 2);

	if (    strlen(xybase) != 2
	     || strchr(VALID_XBASE, xybase[0]) == NULL
	     || strchr(VALID_YBASE, xybase[1]) == NULL )
	{
		return "Invalid base";
	}

	self->wm_xbase = xybase[0];
	self->wm_ybase = xybase[1];
	self->wm_xamount = atoi(rest_av[0]);
	self->wm_yamount = atoi(rest_av[1]);
	if ((iswmt ? 3 : 4) <= ac-1) {
		self->wm_class = strdup(rest_av[2]);
		if (self->wm_class == NULL) return "Out of memory";
	}
	if ((iswmt ? 4 : 5) == ac-1) {
		self->wm_title = strdup(rest_av[3]);
		if (self->wm_title == NULL) return "Out of memory";
	}

	return NULL;
}


static void
_BBCmdWindowMoveCommand_del(BBCmdCommand* _self)
{
	BBCmdWindowMoveCommand* self;

	assert(_self->cmd_descriptor == BBCmdWindowMoveCommandDescriptor);
	self = (BBCmdWindowMoveCommand*)_self;

	free(self->wm_class);
	free(self->wm_title);
}


static int
_BBCmdWindowMoveCommand_execute(const BBCmdCommand* _self)
{
	BBCmdWindowMoveCommand* self;
	int x;
	int y;
	HWND hwnd;
	RECT r;

	assert(_self->cmd_descriptor == BBCmdWindowMoveCommandDescriptor);
	self = (BBCmdWindowMoveCommand*)_self;

	if (self->wm_class == NULL) {
		hwnd = GetForegroundWindow();
		while (!0) {
			HWND t;

			t = GetParent(hwnd);
			if (t == NULL) break;
			hwnd = t;
		}
		hwnd = GetLastActivePopup(hwnd);
	} else {
		hwnd = FindWindow(self->wm_class, self->wm_title);
	}

	if (hwnd == NULL) return BBCmd_RC_FALSE;

	GetWindowRect(hwnd, &r);

	switch (self->wm_xbase) {
	default: assert(0); break;
	case 'W':
		x = self->wm_xamount;
		break;
	case 'E':
		x = (GetSystemMetrics(SM_CXSCREEN) - 1) + self->wm_xamount;
		break;
	case 'R':
		x = r.left + self->wm_xamount;
		break;
	case 'A':
		x = self->wm_xamount;
		break;
	}

	switch (self->wm_ybase) {
	default: assert(0); break;
	case 'N':
		y = self->wm_yamount;
		break;
	case 'S':
		y = (GetSystemMetrics(SM_CYSCREEN) - 1) + self->wm_yamount;
		break;
	case 'R':
		y = r.top + self->wm_yamount;
		break;
	case 'A':
		y = self->wm_yamount;
		break;
	}

	SetWindowPos( hwnd, NULL, x, y, -13, -13,
	              SWP_NOACTIVATE | SWP_NOOWNERZORDER
	                | SWP_NOSIZE | SWP_NOZORDER );

	return BBCmd_RC_TRUE;
}




static const char* _BBCmdWindowMoveCommand_alias[] = {
	"window_move_to",
	NULL
};

const BBCmdCommandDescriptor _BBCmdWindowMoveCommandDescriptor = {
	"window_move",
	_BBCmdWindowMoveCommand_alias,
	"(base, x, y [, class [, title]])\n"
		"Move a window to `x' and `y'.\n"
		"If `class' and/or `title' are omitted,\n"
		"the active window will be moved.\n"
		"`base' is a string which consists of 2 characters,\n"
		"and each character can be one of the following value:\n"
		"\n"
		"N   relative coordinate to top of desktop\n"
		"S   ... to bottom\n"
		"E   ... to right\n"
		"W   ... to left\n"
		"C   ... to center\n"
		"R   ... to current position\n"
		"A   absolute coordinate\n"
		"\n"
		"For example, `WS' means relative coordinate to bottom-left.\n"
	       "`window_move_to(...)' is equal to `window_move('RR', ...)'.",
	sizeof(BBCmdWindowMoveCommand),
	_BBCmdWindowMoveCommand_initialize,
	_BBCmdWindowMoveCommand_del,
	_BBCmdWindowMoveCommand_execute
};




/* ---------------- window_resize ---------------- */

static char
resize_type_from_string(const char* s)
{
	if (!strcmp(s, "resize")) return 'R';
	else if (!strcmp(s, "maximize")) return 'x';
	else if (!strcmp(s, "minimize") || !strcmp(s, "iconize")) return 'n';
	else if (!strcmp(s, "restore")) return 'r';
	else if (!strcmp(s, "shade")) return 's';
	else if (!strcmp(s, "maximize_vertical")) return 'v';
	else if (!strcmp(s, "maximize_horizontal")) return 'h';
	else return '\0';
}

static const char*
_BBCmdWindowResizeCommand_initialize( BBCmdCommand* _self,
                                      int ac, const char* const* av,
                                      const BBCmdCompiler* compiler )
{
	BBCmdWindowResizeCommand* self;

	assert(_self->cmd_descriptor == BBCmdWindowResizeCommandDescriptor);
	self = (BBCmdWindowResizeCommand*)_self;

	self->wr_class = NULL;
	self->wr_title = NULL;

	self->wr_resize_type = resize_type_from_string(av[0]);
	if (self->wr_resize_type == '\0')
	    return "Invalid command name (maybe BBCmd's bug in src/Window.c)";

	if (    ac-1 < 0 + (self->wr_resize_type == 'R' ? 2 : 0)
	     || 2 + (self->wr_resize_type == 'R' ? 2 : 0) < ac-1 )
		return "Incorrect number of arguments";

	if (self->wr_resize_type == 'R') {
		self->wr_width = atoi(av[1]);
		self->wr_height = atoi(av[2]);
		av += 2; 
		ac -= 2;
	}

	if (1 <= ac-1) {
		self->wr_class = strdup(av[1]);
		if (self->wr_class == NULL) return "Out of memory";
	}

	if (2 <= ac-1) {
		self->wr_title = strdup(av[2]);
		if (self->wr_title == NULL) return "Out of memory";
	}

	return NULL;
}


static void
_BBCmdWindowResizeCommand_del(BBCmdCommand* _self)
{
	BBCmdWindowResizeCommand* self;

	assert(_self->cmd_descriptor == BBCmdWindowResizeCommandDescriptor);
	self = (BBCmdWindowResizeCommand*)_self;

	free(self->wr_class);
	free(self->wr_title);
}


static int
_BBCmdWindowResizeCommand_execute(const BBCmdCommand* _self)
{
	BBCmdWindowResizeCommand* self;
	HWND hwnd;
	HWND core;

	assert(_self->cmd_descriptor == BBCmdWindowResizeCommandDescriptor);
	self = (BBCmdWindowResizeCommand*)_self;

	if (self->wr_class == NULL) {
		hwnd = GetForegroundWindow();
		while (!0) {
			HWND t;

			t = GetParent(hwnd);
			if (t == NULL) break;
			hwnd = t;
		}
		hwnd = GetLastActivePopup(hwnd);
	} else {
		hwnd = FindWindow(self->wr_class, self->wr_title);
	}

	if (hwnd == NULL) return BBCmd_RC_FALSE;

	if (self->wr_resize_type == 'R') {
		SetWindowPos( hwnd, NULL, 0, 0,
		              self->wr_width, self->wr_height,
		              SWP_NOACTIVATE | SWP_NOMOVE
			          | SWP_NOOWNERZORDER | SWP_NOZORDER );
	} else {
		core = BBCmd_FindCore();

		if (core != NULL) {
			UINT msg;
			WPARAM wp = 0;
			LPARAM lp = 0;

			switch (self->wr_resize_type) {
			default: assert(0); break;
			case 'x': msg = BB_WINDOWMAXIMIZE;
			          lp = (LPARAM)hwnd; break;
			case 'n': msg = BB_WINDOWMINIMIZE;
			          lp = (LPARAM)hwnd; break;
			case 'r': msg = BB_WINDOWRESTORE;
			          lp = (LPARAM)hwnd; break;
			case 's': msg = BB_WINDOWSHADE;
			          wp = (WPARAM)hwnd; break;
			case 'v': msg = BB_WINDOWGROWHEIGHT;
			          wp = (WPARAM)hwnd; break;
			case 'h': msg = BB_WINDOWGROWWIDTH;
			          wp = (WPARAM)hwnd; break;
			}

			PostMessage(core, msg, wp, lp);
		} else {
			WPARAM wp;

			switch (self->wr_resize_type) {
			default: assert(0); break;
			case 'x': wp = SC_MAXIMIZE; break;
			case 'n': wp = SC_MINIMIZE; break;
			case 'r': wp = SC_RESTORE; break;
			case 's': /* FALLTHRU */
			case 'v': /* FALLTHRU */
			case 'h': return BBCmd_RC_FALSE;
			}

			PostMessage(hwnd, WM_SYSCOMMAND, wp, 0);
		}
	}

	return BBCmd_RC_TRUE;
}




static const char* _BBCmdWindowResizeCommand_alias[] = {
	"maximize",
	"minimize",
	"iconize",
	"restore",
	"shade",
	"maximize_vertical",
	"maximize_horizontal",
	NULL
};

const BBCmdCommandDescriptor _BBCmdWindowResizeCommandDescriptor = {
	"resize",
	_BBCmdWindowResizeCommand_alias,
	"(width, height [, class [, title]])\n"
		"Resize the specified window.\n"
		"These commands will work for the active window\n"
		"if `class' and `title' are not given.\n"
		"`maximize([class [, title]])' will maximize,\n"
		"`minimize([class [, title]])' will minimize, and so on.\n"
		"Note that `maximize_vertical', `maximize_horizontal'\n"
		"and `shade' does not work if the core is not running.",
	sizeof(BBCmdWindowResizeCommand),
	_BBCmdWindowResizeCommand_initialize,
	_BBCmdWindowResizeCommand_del,
	_BBCmdWindowResizeCommand_execute
};




/* __END__ */
